--
-- OrganizedPanic
--
-- Copyright (c) 2021 Dana Basken
--

local _, NS = ...

local VERSION="9.1.0.16"

OrganizedPanic = LibStub("AceAddon-3.0"):NewAddon(NS.ADDON_NAME, "AceConsole-3.0", "AceTimer-3.0", "AceEvent-3.0", "AceComm-3.0")

print(OrganizedPanicCommon:PrefixString(NS.ADDON_NAME, ("v%s"):format(VERSION), NS.PREFIX_COLOR))

local OPTIONS = NS:BuildOptions(NS.ADDON_NAME, OrganizedPanic)

local AceGUI = LibStub("AceGUI-3.0") -- needed for AceConfigDialog
local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale(NS.ADDON_NAME)

local COOLDOWN_UPDATE_INTERVAL = 3
local UPDATE_SPELL_COOLDOWNS_INTERVAL = 0.5
local SEND_VERSION_MESSAGE_INTERVAL = 300
local INSTANT_SPELL_AUTO_STOP_TIMEOUT = 0.5

function OrganizedPanic:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New(NS.DB_NAME, NS.DEFAULTS, true)

  NS.db = self.db
  NS.debug = OrganizedPanic.debug
  NS.warn = OrganizedPanic.warn
  NS.trace = OrganizedPanic.trace
  NS.stack = OrganizedPanic.stack
  NS.ObjectToString = OrganizedPanicCommon.ObjectToString
  NS.VERSION = VERSION

  self.minimapIcon = OrganizedPanicMinimapIcon:Create(("%sMinimap"):format(NS.ADDON_NAME), "Interface/AddOns/OrganizedPanic/images/op_icon_normal.tga", self.db, function() self.window:Toggle() end)

  LibStub("AceConfig-3.0"):RegisterOptionsTable(NS.ADDON_NAME, OPTIONS)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(NS.ADDON_NAME)

  self.flasher = BorderFlasher:Create("Interface\\Addons\\OrganizedPanic\\images\\border.tga", self.db.profile.borderColor)
  self.castingStates = {}
  self.latestVersionSeen = VERSION
  self.members = OrganizedPanicMembers:Create()
  self.window = OrganizedPanicWindow:Create(self.members)

  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_START", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_STOP", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "SPELLCAST_EVENT_HANDLER")

  OrganizedPanic:RegisterEvent("GROUP_ROSTER_UPDATE", "GROUP_UPDATE_HANDLER")
  OrganizedPanic:RegisterEvent("GROUP_JOINED", "GROUP_UPDATE_HANDLER")
  OrganizedPanic:RegisterEvent("GROUP_LEFT", "GROUP_UPDATE_HANDLER")
  OrganizedPanic:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")

  AceEvent.RegisterMessage(self, "CASTING_STATE_CHANGE")
  AceEvent.RegisterMessage(self, "SPELL_STATE_CHANGE")
  AceEvent.RegisterMessage(self, "SPELL_COOLDOWN")
  AceEvent.RegisterMessage(self, "MEMBER_INSPECTED")
  AceEvent.RegisterMessage(self, "AURA_STATE_CHANGE")
  AceEvent.RegisterMessage(self, "GROUPED_STATE_CHANGE")

  OrganizedPanic:RegisterComm(NS.ADDON_NAME)
  OrganizedPanic:RegisterChatCommand(NS.CHAT_COMMAND_NAME, "HandleChatCommand")
  OrganizedPanic:RegisterChatCommand(NS.CHAT_ALT_COMMAND_NAME, "HandleChatCommand")
  self:UpdateWindowVisibility()
end

function OrganizedPanic:OnEnable()
  AceTimer.ScheduleRepeatingTimer(self, "UpdateSpellCooldowns", UPDATE_SPELL_COOLDOWNS_INTERVAL)
  AceTimer.ScheduleRepeatingTimer(self, "SendVersionMessages", SEND_VERSION_MESSAGE_INTERVAL)
  self:GROUP_UPDATE_HANDLER()
end

function OrganizedPanic:OnDisable()
  AceTimer:CancelAllTimers()
end

function OrganizedPanic:PLAYER_SPECIALIZATION_CHANGED(event, unitId)
  local member = self.members:FindMemberByGroupName(unitId)
  NS:debug(event, unitId, member)
  if member then member:SetNeedsInspect() end
  self.members:Update()
  self:SendGroupAddonMessage("spec_change")
end

function OrganizedPanic:GROUP_UPDATE_HANDLER(...)
  NS:debug("GROUP_UPDATE_HANDLER", ...)
  self.members:Update()
end

function OrganizedPanic:MEMBER_INSPECTED(event, name, unitName)
  self:SendGroupAddonMessage("send_covenant")
  self:SendGroupAddonMessage("send_spells")
end

function OrganizedPanic:CASTING_STATE_CHANGE(event, name, state, spellId)
  NS:trace(event, name, state, spellId)
  local member = self.members:FindMemberByName(name)
  if not member then return end
  if state == "SPELL_START" then
    local spell = member:FindMemberSpellById(spellId)
    if spell then member:SetSpellStart(spell, GetTime()) end -- set initial start, will get updated if member has addon
    local caster = OrganizedPanicCommon:GetNameWithoutRealm(member:GetName())
    if self.db.profile.includeRole then
      if UnitGroupRolesAssigned("player") == member:GetRole() then self:Panic(caster, spellId) end
    else
      self:Panic(caster, spellId)
    end
  end
end

function OrganizedPanic:SPELL_STATE_CHANGE(event, name, spellId, state)
  NS:trace(event, name, spellId, state)
  local member = self.members:FindMemberByName(name)
  if not member then return end
  if member:IsPlayer() then
    local spell = member:FindMemberSpellById(spellId)
    if spell then self:SendSpellMessage(member, spell) end
  end
end

function OrganizedPanic:SPELL_COOLDOWN(event, name, spellId, remaining)
  NS:trace(event, name, spellId, remaining)
  local member = self.members:FindMemberByName(name)
  if not member then return end
  if member:IsPlayer() then
    local now = GetTime()
    if not self.lastSpellCooldownSent then self.lastSpellCooldownSent = GetTime() end
    if now - self.lastSpellCooldownSent <= COOLDOWN_UPDATE_INTERVAL then return end
    local spell = member:FindMemberSpellById(spellId)
    if spell then
      self:SendSpellMessage(member, spell)
      self.lastSpellCooldownSent = now
    end
  end
end

function OrganizedPanic:AURA_STATE_CHANGE(event, name, spellId, state, casterGroupName)
  NS:debug(event, name, spellId, state, casterGroupName)
  if not NS.AURAS[spellId] then return end
  local member = self.members:FindMemberByName(name)
  if not member then return end
  local caster = self.members:FindMemberByGroupName(casterGroupName)
  if not caster then return end
  if state == "ON" then self:Buff(OrganizedPanicCommon:GetNameWithoutRealm(caster.name), OrganizedPanicCommon:GetNameWithoutRealm(member.name), spellId) end
end

function OrganizedPanic:GROUPED_STATE_CHANGE(event, grouped)
  self:UpdateWindowVisibility()
end

function OrganizedPanic:UpdateWindowVisibility()
  if NS.db.profile.autoShowHide then
    if self.members:GetGroupedState() then
      self.window:Show()
    else
      self.window:Hide()
    end
  end
end

-- every 0.5 seconds
function OrganizedPanic:UpdateSpellCooldowns()
  if not OrganizedPanicCommon:InGlobalCooldown() then
    self:UpdatePlayerSpellCooldowns()
  end
end

-- every 0.5 seconds, if not in GCD
function OrganizedPanic:UpdatePlayerSpellCooldowns()
  local member = self.members:FindPlayer()
  if not member then return end
  for _, spell in pairs(member:GetSpells()) do
    local start, duration, enabled = GetSpellCooldown(spell.id);
    if enabled == 0 then -- spell active, like Stealth
      -- don't think there's anything to do
    elseif start > 0 and duration > 0 then -- cooling down
      member:SetSpellStart(spell, start, duration)
    else -- ready
      member:SetSpellStart(spell, 0)
    end
  end
end

function OrganizedPanic:SendMySpells(name)
  local member = self.members:FindPlayer()
  if not member then return end
  local hasSpells = false
  for _, spell in pairs(member.spells) do
    hasSpells = true
    self:SendSpellMessage(member, spell, name)
  end
  if not hasSpells then self:SendGroupAddonMessage("no_spells") end
end

function OrganizedPanic:SendMyCovenant(name)
  local member = self.members:FindPlayer()
  if not member then return end
  local message = ("covenant:%s:%d"):format(member.name, C_Covenants.GetActiveCovenantID())
  self:SendGroupAddonMessage(message)
end

function OrganizedPanic:SendSpellMessage(member, spell, name)
  local start = math.floor(spell.start)
  if start > 0 then
    start = math.floor(GetTime() - spell.start)
  end
  local remaining = math.floor(spell.remaining)
  local duration = math.floor(spell.duration)
  local message = ("spell:%s:%s:%s:%d:%d:%d"):format(member.name, spell.id, spell.state, start, remaining, duration)
  self:SendGroupAddonMessage(message)
end

function OrganizedPanic:SendGroupAddonMessage(message, priority)
  local channel = OrganizedPanicCommon:GetChannel()
  if not channel then return end
  self:SendCommMessageInternal(message, channel, nil, priority)
end

function OrganizedPanic:SendVersionMessages()
  local channel = OrganizedPanicCommon:GetChannel()
  if channel then
    self:SendVersionMessage(channel)
  else
    if IsInGuild() then self:SendVersionMessage("GUILD") end
  end
end

function OrganizedPanic:SendVersionMessage(channel)
  self:SendCommMessageInternal(("version:%s"):format(self.latestVersionSeen), channel)
end

function OrganizedPanic:SendCommMessageInternal(message, channel, name, priority)
  NS:debug("SendCommMessage", message, channel, name, priority)
  if name and name == OrganizedPanicCommon:GetPlayerNameWithRealm() then
    NS:debug("SendCommMessage", "not sending to myself")
    return
  end
  self:SendCommMessage(NS.ADDON_NAME, message, channel, name, priority)
end

-- handle all API spell events, and turn them into SPELL_START or SPELL_STOP, and have member handle the state change
function OrganizedPanic:SPELLCAST_EVENT_HANDLER(event, unit, guid, spellId)
  if not NS.db.profile.standalone and not OrganizedPanicCommon:InGroup() then return end
  -- find member
  local member = self.members:FindMemberByUnitName(unit)
  if not member then return end
  -- convert API event to simple start/stop event
  if event == "UNIT_SPELLCAST_SUCCEEDED" or event == "UNIT_SPELLCAST_START" or event == "UNIT_SPELLCAST_CHANNEL_START" then event = "SPELL_START" end
  if event == "UNIT_SPELLCAST_STOP" or event == "UNIT_SPELLCAST_CHANNEL_STOP" then event = "SPELL_STOP" end
  -- apply spell modifications
  if event == "SPELL_START" then member:ApplySpellModifications(spellId) end
   -- if this is a panic spell, update spell state
  local spell = NS.PANIC_SPELLS[spellId]
  if not spell then return end
  member:SetCastingState(event, spellId)
  if spell.type == "INSTANT" then
    C_Timer.After(INSTANT_SPELL_AUTO_STOP_TIMEOUT, function() member:SetCastingState("SPELL_STOP", spellId) end)
  end
  if spell.type == "SHIFT" then
    C_Timer.After(spell.shift_duration, function() member:SetCastingState("SPELL_STOP", spellId) end)
  end
end

function OrganizedPanic:Panic(caster, id, force)
  if not force then
    if not self.db.profile.includeSelf and caster == OrganizedPanicCommon:GetPlayerNameWithoutRealm() then return end
  end
  self:PlayAlertSound()
  self:ShowRaidWarning(caster, id)
  self:FlashBorder()
end

function OrganizedPanic:PlayAlertSound()
  if self.db.profile.playSound then self:PlaySound(NS.SOUNDS[self.db.profile.sound]) end
end

function OrganizedPanic:PlaySound(sound)
  if sound.id then
    NS:debug("PlaySound", "id", sound.id)
    PlaySound(sound.id, "Master")
  elseif sound.path then
    NS:debug("PlaySound", "path", sound.path)
    PlaySoundFile(sound.path, "Master")
  end
end

function OrganizedPanic:ShowRaidWarning(caster, id)
  if not caster then caster = NS.TEST_PANIC_CASTERS[math.random(#NS.TEST_PANIC_CASTERS)] end
  if not id then id = NS.TEST_PANIC_SPELLS[math.random(#NS.TEST_PANIC_SPELLS)] end
  local name, _ = GetSpellInfo(id)
  local message = (L["panic_template"]):format(caster, name)
  RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])
end

function OrganizedPanic:Buff(caster, target, spellId, force)
  NS:debug("Buff", caster, target, spellId, force)
  if not force then
    if not self.db.profile.includeSelf and caster == OrganizedPanicCommon:GetPlayerNameWithoutRealm() then return end
  end
  self:PlayBuffAlertSound()
  self:ShowRaidBuffWarning(caster, target, spellId)
end

function OrganizedPanic:PlayBuffAlertSound()
  if self.db.profile.playSound then self:PlaySound(NS.SOUNDS[self.db.profile.buffSound]) end
end

function OrganizedPanic:ShowRaidBuffWarning(caster, target, spellId)
  if not caster then caster = NS.TEST_PANIC_CASTERS[math.random(#NS.TEST_PANIC_CASTERS)] end
  if not target then target = NS.TEST_PANIC_CASTERS[math.random(#NS.TEST_PANIC_CASTERS)] end
  if not spellId then spellId = NS.TEST_PANIC_SPELLS[math.random(#NS.TEST_PANIC_SPELLS)] end
  local name, _ = GetSpellInfo(spellId)
  local message = (L["buff_template"]):format(caster, target, name)
  if caster == target then
    message = (L["self_buff_template"]):format(caster, name)
  end
  RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])
end

function OrganizedPanic:FlashBorder()
  if self.db.profile.flashBorder then self.flasher:StartFlashing() end
end

function OrganizedPanic:GetOption(info)
  local property = info.option.arg
  return self.db.profile[property]
end

function OrganizedPanic:SetOption(info, value)
  local property = info.option.arg
  self.db.profile[property] = value
end

function OrganizedPanic:GetMimimapVisible(info)
  return not self.db.profile.minimap.hide
end

function OrganizedPanic:SetMimimapVisible(info, value)
  self.db.profile.minimap.hide = not value
  self.minimapIcon:UpdateVisibility()
end

function OrganizedPanic:SetBorderColor(info, r, g, b, a)
  self.db.profile.borderColor.r = r
  self.db.profile.borderColor.g = g
  self.db.profile.borderColor.b = b
  self.flasher:SetBorderColor(self.db.profile.borderColor)
  self.flasher:StartFlashing()
end

function OrganizedPanic:GetBorderColor(info)
  return self.db.profile.borderColor.r, self.db.profile.borderColor.g, self.db.profile.borderColor.b, 1.0
end

function OrganizedPanic:AddonPrint(format, ...)
  print(OrganizedPanicCommon:PrefixString(NS.ADDON_NAME, string.format(format, ...), NS.PREFIX_COLOR))
end

function OrganizedPanic:HandleChatCommand(input)
  if not input or input:trim() == "" then
    LibStub("AceConfigDialog-3.0"):Open(NS.ADDON_NAME)
  else
    if input == "help" then
      self:AddonPrint("%s", L["Commands:"])
      self:AddonPrint(("  /op minimap - %s"):format(L["toggle minimap icon"]))
      self:AddonPrint(("  /op hide - %s"):format(L["hide window"]))
      self:AddonPrint(("  /op show - %s"):format(L["show window"]))
      self:AddonPrint(("  /op lock - %s"):format(L["lock window"]))
      self:AddonPrint(("  /op unlock - %s"):format(L["unlock window"]))
    elseif input == "panic" then
      self:Panic(nil, nil, true)
    elseif input == "buff" then
      self:Buff()
    elseif input == "settings" then
      self:ShowSettings()
    elseif input == "debug" then
      self.db.profile.loglevel = NS.LOG_LEVEL.DEBUG
      self:ShowSettings()
    elseif input == "warn" then
      self.db.profile.loglevel = NS.LOG_LEVEL.WARN
      self:ShowSettings()
    elseif input == "trace" then
      self.db.profile.loglevel = NS.LOG_LEVEL.TRACE
      self:ShowSettings()
    elseif input == "stack" then
      self.db.profile.loglevel = NS.LOG_LEVEL.STACK
      self:ShowSettings()
    elseif input == "normal" then
      self.db.profile.loglevel = NS.LOG_LEVEL.NONE
      self.db.profile.standalone = false
      self.members:Update()
      self:ShowSettings()
    elseif input == "standalone" then
      if self.db.profile.standalone then self.db.profile.standalone = false else self.db.profile.standalone = true end
      self.members:Update()
      self:ShowSettings()
    elseif input == "auto" then
      if self.db.profile.autoShowHide then self.db.profile.autoShowHide = false else self.db.profile.autoShowHide = true end
      self.members:Update()
      self:ShowSettings()
    elseif input == "members" then
      for _, member in pairs(self.members:GetMembers()) do
        print(member.name, member:GetClass(), member:GetSpecialization(), member:GetRole(), member:GetGroupName())
        for _, spell in pairs(member:GetSpells()) do
          print("  ", spell.name, spell.id, spell.state, spell.start, spell.duration, spell.remaining)
        end
      end
    elseif input == "dump" then
      for _, member in pairs(self.members:GetMembers()) do
        print(NS:ObjectToString(member))
      end
    elseif input == "minimap" then
      self.minimapIcon:Toggle()
    elseif input == "toggle" then
      self.window:Toggle()
    elseif input == "show" then
      self.window:Show()
    elseif input == "hide" then
      self.window:Hide()
    elseif input == "lock" then
      self.window:Lock()
    elseif input == "unlock" then
      self.window:Unlock()
    else
      self:AddonPrint("unknown command: '%s', try '/%s help'", input, NS.CHAT_COMMAND_NAME)
    end
  end
end

function OrganizedPanic:ShowSettings()
  self:AddonPrint("SETTINGS:")
  self:AddonPrint("  loglevel: %d", self.db.profile.loglevel)
  self:AddonPrint("  includeRole: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.includeRole))
  self:AddonPrint("  includeSelf: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.includeSelf))
  self:AddonPrint("  standalone: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.standalone))
  self:AddonPrint("  flashBorder: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.flashBorder))
  self:AddonPrint("  autoShowHide: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.autoShowHide))
  self:AddonPrint("  sound: %s", self.db.profile.sound)
  self:AddonPrint("  playSound: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.playSound))
end

function OrganizedPanic:HandleSpellMessage(data)
  local name, spellId, state, start, remaining, duration = strsplit(":", data)
  local member = self.members:FindMemberByName(name)
  if not member then
    NS:warn("HandleSpellMessage", "could not find member", name)
    return
  end
  local spell = member:FindMemberSpellById(tonumber(spellId))
  if not spell then
    NS:warn("HandleSpellMessage", "could not find spell", member.name, spellId)
    return
  end
  start = tonumber(start)
  duration = tonumber(duration)
  if state == "COOLDOWN" then
    local now = GetTime()
    start = math.floor(now - start)
    if start < 0 then -- fix users using 9.1.0.9 or before that send raw GetTime()
      start = math.floor(now + tonumber(remaining) - duration)
      NS:warn("HandleSpellMessage", "fix broken start time", member.name, spell.name)
    end
  else
    start = 0
  end
  member:SetSpellStart(spell, start, duration, state)
end

function OrganizedPanic:HandleSpecChangeMessage(name)
  local member = self.members:FindMemberByName(name)
  if not member then
    NS:warn("HandleSpecChangeMessage", "could not find member", name)
    return
  end
  NS:debug("HandleSpecChangeMessage", member.name)
  member:SetNeedsInspect()
end

function OrganizedPanic:HandleCovenantMessage(data)
  local name, covenant = strsplit(":", data)
  local member = self.members:FindMemberByName(name)
  if not member then
    NS:warn("HandleSpecChangeMessage", "could not find member", name)
    return
  end
  member:SetCovenant(tonumber(covenant))
end

function OrganizedPanic:HandleVersionMessage(version)
  if OrganizedPanicCommon:IsNewerVersion(version, self.latestVersionSeen) then
    self:AddonPrint(L["new_version_template"], version)
    self.latestVersionSeen = version
  end
end

function OrganizedPanic:OnCommReceived(prefix, message, distribution, sender)
  local event, data = strsplit(":", message, 2)
  local sender = OrganizedPanicCommon:GetNameWithRealm(sender)
  NS:debug("OnCommReceived", event, data, sender, distribution)
  if sender ~= OrganizedPanicCommon:GetPlayerNameWithRealm() then
    if event == "version" then self:HandleVersionMessage(data) end
    if event == "send_spells" then self:SendMySpells(sender) end
    if event == "send_covenant" then self:SendMyCovenant(sender) end
    if event == "spell" then self:HandleSpellMessage(data) end
    if event == "covenant" then self:HandleCovenantMessage(data) end
    if event == "spec_change" then self:HandleSpecChangeMessage(sender) end
    local member = self.members:FindMemberByName(sender)
    if member then member:SetHasAddon() end
  end
end

function OrganizedPanic:GetSoundOptions()
  local sounds = {}
  for id, sound in pairs(NS.SOUNDS) do sounds[id] = sound.name end
  return sounds
end

function OrganizedPanic:GetSoundOptionsSort()
  local keys = {}
  for key, sound in pairs(NS.SOUNDS) do table.insert(keys, key) end
  table.sort(keys, function(a, b)
    if NS.SOUNDS[a].top then return true end
    if NS.SOUNDS[b].top then return false end
    return NS.SOUNDS[a].name < NS.SOUNDS[b].name
  end)
  return keys
end

function OrganizedPanic:debug(where, ...)
  if NS.db.profile.loglevel >= NS.LOG_LEVEL.DEBUG then
    print(OrganizedPanicCommon:ColorString("op-debug:", "8080FF"), OrganizedPanicCommon:ColorString(where, "808080"), ...)
    NS:stack(where)
  end
end

function OrganizedPanic:warn(where, ...)
  if NS.db.profile.loglevel >= NS.LOG_LEVEL.WARN then
    print(OrganizedPanicCommon:ColorString("op-warn:", "FC5E03"), OrganizedPanicCommon:ColorString(where, "808080"), ...)
    NS:stack(where)
  end
end

function OrganizedPanic:trace(where, ...)
  if NS.db.profile.loglevel >= NS.LOG_LEVEL.TRACE then
    print(OrganizedPanicCommon:ColorString("op-trace:", "FF80FF"), OrganizedPanicCommon:ColorString(where, "808080"), ...)
    NS:stack(where)
  end
end

function OrganizedPanic:stack(where)
  if NS.db.profile.loglevel >= NS.LOG_LEVEL.STACK then
    print(OrganizedPanicCommon:ColorString("op-stack:", "FF80FF"), OrganizedPanicCommon:ColorString(where, "808080"), debugstack(2))
  end
end
