--
-- OrganizedPanicMember
--
-- Copyright (c) 2021 Dana Basken
--

local _, NS = ...

local AceEvent = LibStub("AceEvent-3.0")

local SPELL_MODIFICATION_COOLDOWN = 10
local MAX_AURAS = 40

OrganizedPanicMember = {}
OrganizedPanicMember.__index = OrganizedPanicMember

function OrganizedPanicMember:Create(name)
  local instance = {}
  setmetatable(instance, OrganizedPanicMember)
  instance:Construct(name)
  return instance
end

function OrganizedPanicMember:Construct(name)
  self.name = name
  self.isPlayer = UnitName("player") == self:GetName(true)
  self.hasAddon = self.isPlayer
  self.lastInspectedTime = 0
  self.spellModificationTimes = {}
  self.auras = {
    {state = "OFF", name = "HasAddon", icon = "Interface/AddOns/OrganizedPanic/images/op_icon_normal.tga", spellId = "HasAddon", source = "player", fake = true},
    {state = "OFF", name = "NeedsInspect", icon = 134442, spellId = "NeedsInspect", source = "player", fake = true},
  }
  self:Update()
end

function OrganizedPanicMember:Update()
  self:UpdateCovenant()
  self:UpdateGroupType()
  self:UpdateGUID()
  self:UpdateUnitName()
  self:UpdateRole()
  self:UpdateClass()
  self:UpdateTalents()
end

-- function OrganizedPanicMember:__tostring()
--   return ("OrganizedPanicMember: %s"):format(NS:ObjectToString(self))
-- end

function OrganizedPanicMember:SetSpecialization(specialization)
  NS:debug("SetSpecialization", self.name, NS:ObjectToString(specialization))
  self.specialization = specialization
  self:UpdateSpells()
end

function OrganizedPanicMember:SetSpecializationIcon(icon)
  self.specializationIcon = icon
end

function OrganizedPanicMember:SetRole(role)
  self.role = role
  self:UpdateSpells()
end

function OrganizedPanicMember:SetClass(class)
  self.class = class
  self:UpdateSpells()
end

function OrganizedPanicMember:SetTalents(talents)
  self.talents = talents
  self:UpdateSpells()
end

function OrganizedPanicMember:SetSpells(spells)
  self.spells = spells
end

function OrganizedPanicMember:SetGroupName(groupName)
  self.groupName = groupName
end

function OrganizedPanicMember:SetCovenant(covenant)
  self.covenant = covenant
  self:UpdateSpells()
end

-- this is called from OrganizedPanic with the current state, decide if state has changed and sent CASTING_STATE_CHANGE if so
function OrganizedPanicMember:SetCastingState(event, spellId)
  local spell = self:FindMemberSpellById(spellId)
  if not spell then return end
  local now = GetTime()
  if not self.castingState then self.castingState = {event = nil, spellId = nil, when = now} end
  if self.castingState.spellId ~= spellId or self.castingState.event ~= event or now - self.castingState.when > 30 then
    self.castingState = {event = event, spellId = spellId, when = now}
    AceEvent:SendMessage("CASTING_STATE_CHANGE", self.name, event, spellId)
  end
end

function OrganizedPanicMember:UpdateRole()
  local role = UnitGroupRolesAssigned(self.name)
  if not role or role == "NONE" then
    name, _ = strsplit("-", self.name)
    role = UnitGroupRolesAssigned(name)
  end
  self:SetRole(role)
end

function OrganizedPanicMember:UpdateClass()
  local _, class, _ = UnitClass(self.name);
  if not class then
    local name, _ = strsplit("-", self.name)
    _, class, _ = UnitClass(name);
  end
  self:SetClass(class)
end

function OrganizedPanicMember:UpdateSpells()
  local spells = {}
  for id, panicSpell in pairs(NS.PANIC_SPELLS) do
    if self:HasPanicSpell(panicSpell) then
      NS:debug("UpdateSpells", self.name, NS:ObjectToString(panicSpell))
      local spell = self:FindMemberSpellById(id)
      if not spell then spell = self:CloneSpell(panicSpell, id) end
      table.insert(spells, spell)
    end
  end
  self:SetSpells(spells)
end

function OrganizedPanicMember:HasPanicSpell(panicSpell)
  if panicSpell.class ~= self.class then return false end
  if panicSpell.specialization ~= self.specialization then return false end
  if panicSpell.talentId and not self:HasTalentById(panicSpell.talentId) then return false end
  if panicSpell.covenantId and panicSpell.covenantId ~= self.covenant then return false end
  return true
end

function OrganizedPanicMember:CloneSpell(spell, id)
  NS:debug("CloneSpell", self.name, NS:ObjectToString(spell))
  local clone = {}
  for k, v in pairs(spell) do clone[k] = v end
  clone.id = id
  clone.remaining = 0
  clone.start = 0
  clone.state = spell.state or "READY"
  clone.name = GetSpellInfo(id)
  self:ApplyTalentModifications(clone)
  NS:trace("CloneSpell", self.name, "clone", NS:ObjectToString(clone))
  return clone
end

function OrganizedPanicMember:ApplyTalentModifications(spell)
  for _, modification in pairs(self:GetTalentModifications(spell.id)) do
    if modification.type == "offset" then
      spell[modification.property] = spell[modification.property] + modification.value
      NS:debug("ApplyTalentModifications", self.name, NS:ObjectToString(modification), NS:ObjectToString(spell))
    end
  end
end

function OrganizedPanicMember:GetTalentModifications(spellId)
  NS:debug("GetTalentModifications", self.name, spellId)
  local modifications = {}
  for talentId, modification in pairs(NS.TALENT_MODIFICATIONS) do
    if self:HasTalentById(modification.talentId) and modification.spellId == spellId then table.insert(modifications, modification) end
  end
  return modifications
end

function OrganizedPanicMember:HasTalentById(talentId)
  for _, talent in pairs(self.talents) do
    if talent.spellId == talentId then return true end
  end
  return false
end

function OrganizedPanicMember:ApplySpellModifications(spellId)
  local modification = NS.SPELL_MODIFICATIONS[spellId]
  if modification then
    local now = GetTime()
    if not self.spellModificationTimes[spellId] then self.spellModificationTimes[spellId] = 0 end
    if now - self.spellModificationTimes[spellId] < SPELL_MODIFICATION_COOLDOWN then return end
    local spell = self:FindMemberSpellById(modification.spellId)
    if not spell then return end
    NS:debug("ApplySpellModifications", self.name, spellId, spell.name, NS:ObjectToString(NS.SPELL_MODIFICATIONS[spellId]))
    if modification.type == "offset" then
      spell[modification.property] = spell[modification.property] + modification.value
    end
    self:UpdateSpellCooldowns()
    self.spellModificationTimes[spellId] = now
  end
end

function OrganizedPanicMember:UpdateGUID()
  self.guid = UnitGUID(self.name)
  if not self.guid then
    local name, realm = strsplit("-", self.name)
    self.guid = UnitGUID(name)
  end
end

-- actual player name, with or without realm
function OrganizedPanicMember:GetName(short)
  if short then
    local name, realm = strsplit("-", self.name)
    return name
  end
  return self.name
end

-- player GUID, like Player-64-0C49D7E4
function OrganizedPanicMember:GetGUID()
  return self.guid
end

-- group name, like party1
function OrganizedPanicMember:GetGroupName()
  self:UpdateGroupName()
  return self.groupName
end

-- could be name-realm or name, depending on what works for UnitName()
function OrganizedPanicMember:GetUnitName()
  self:UpdateUnitName()
  return self.unitName
end

function OrganizedPanicMember:GetRole()
  return self.role
end

function OrganizedPanicMember:GetSpells()
  return self.spells or {}
end

function OrganizedPanicMember:GetClass()
  return self.class
end

function OrganizedPanicMember:GetSpecialization()
  return self.specialization
end

function OrganizedPanicMember:GetAuras()
  return self.auras
end

function OrganizedPanicMember:FindMemberSpellById(id)
  for _, spell in pairs(self:GetSpells()) do
    if spell.id == id then return spell end
  end
end

function OrganizedPanicMember:UpdateGroupType()
  self.groupType = nil
  local unitName = self:GetUnitName()
  if UnitInParty(unitName) then self.groupType = "party" end
  if UnitInRaid(unitName) then self.groupType = "raid" end
end

function OrganizedPanicMember:IsPlayer()
  return self.isPlayer
end

function OrganizedPanicMember:UpdateUnitName()
  self.unitName = UnitName(self.name)
  if not self.unitName then
    local name, realm = strsplit("-", self.name)
    self.unitName = UnitName(name)
  end
end

function OrganizedPanicMember:UpdateGroupName()
  if self.groupType then
    local members = GetNumGroupMembers()
    local groupName = nil
    for i = 1, members do
      local playerGroupName = ("%s%d"):format(self.groupType, i)
      local guid = UnitGUID(playerGroupName)
      if not guid then
        playerGroupName = "player"
        guid = UnitGUID(playerGroupName)
      end
      if guid == self.guid then self:SetGroupName(playerGroupName) end
    end
  else
    self:SetGroupName(nil)
  end
end

function OrganizedPanicMember:UpdateTalents()
  local talents = {}
  local playerGroupName = nil;
  if not self:IsPlayer() then playerGroupName = self:GetGroupName() end
  local inspect = playerGroupName ~= nil
  for r = 1, MAX_TALENT_TIERS do
    for c = 1, NUM_TALENT_COLUMNS do
      local id, name, texture, selected, available, spellId = GetTalentInfo(r, c, 1, inspect, playerGroupName)
      local talent = {spellId = spellId, name = name}
      if available and selected then table.insert(talents, talent) end
    end
  end
  if #talents then
    NS:debug("UpdateTalents", self.name, NS:ObjectToString(talents))
    self:SetTalents(talents)
  else
    NS:debug("UpdateTalents", self.name, "empty talents, not setting")
  end
end

function OrganizedPanicMember:UpdateCovenant()
  if self:IsPlayer() then
    self.covenant = C_Covenants.GetActiveCovenantID()
  end
end

-- every second, only for player, and if not in GCD (also called in response to "spell" channel message)
function OrganizedPanicMember:SetSpellStart(spell, start, duration, state)
  -- NS:trace("SetSpellStart", self.name, spell.name, start, duration, state, spell.state)
  spell.start = math.floor(start)
  if duration and duration > 0 then
    spell.duration = math.floor(duration)
  end
  if state then
    self:SetSpellState(spell, state)
  end
end

-- every second, from OrganizedPanicMembers:UpdateCooldownsRemaining()
function OrganizedPanicMember:UpdateSpellCooldowns()
  for _, spell in pairs(self:GetSpells()) do
    if spell.start and spell.start > 0 then
      local now = math.floor(GetTime())
      spell.remaining = math.floor(spell.duration - (now - spell.start))
    else
      spell.remaining = 0
    end
    if spell.remaining > 0 then
      self:SetSpellState(spell, "COOLDOWN")
      AceEvent:SendMessage("SPELL_COOLDOWN", self.name, spell.id, spell.remaining)
    else
      spell.remaining = 0
      self:SetSpellState(spell, "READY")
    end
  end
end

function OrganizedPanicMember:SetSpellState(spell, state)
  if spell.state ~= state then
    NS:debug("SetSpellState", self.name, spell.name, state)
    spell.state = state
    AceEvent:SendMessage("SPELL_STATE_CHANGE", self.name, spell.id, spell.state)
  end
end

function OrganizedPanicMember:SetHasAddon()
  if self.hasAddon ~= true then
    self.hasAddon = true
    AceEvent:SendMessage("MEMBER_UPDATED", self.name)
    NS:debug("SetHasAddon", self.name, self.hasAddon)
  end
end

function OrganizedPanicMember:HasAddon()
  return self.hasAddon
end

function OrganizedPanicMember:NeedsInspect()
  if not self.talents or #self.talents == 0 then return true end
  if not self.specialization then return true end
  return false
end

function OrganizedPanicMember:SetNeedsInspect()
  self:SetTalents({})
  self:SetSpecialization(nil)
end

function OrganizedPanicMember:SetInspecting()
  self.lastInspectedTime = GetTime()
end

function OrganizedPanicMember:UpdateAuras()
  local groupName = self:GetGroupName()
  if groupName then
    local seen = {}
    for i = 1, MAX_AURAS do
      local name, icon, count, _, duration, expiration, source, _, _, spellId, _, _, castByPlayer = UnitAura(groupName, i)
      if not name then break end
      if NS.AURAS[spellId] then
        spellId = NS.AURAS[spellId].spellId
        seen[spellId] = true
        local aura = self.auras[spellId]
        if not aura then
          aura = {name = name, icon = icon, duration = duration, expiration = expiration, spellId = spellId}
          self.auras[spellId] = aura
        end
        aura.source = source
        self:SetAuraState(aura, "ON")
      end
    end
    for _, aura in pairs(self.auras) do
      if aura.fake then
        self:UpdateFakeAura(aura)
      else
        if not seen[aura.spellId] then self:SetAuraState(aura, "OFF") end
      end
    end
  end
  return self.auras
end

function OrganizedPanicMember:UpdateFakeAura(aura)
  if aura.spellId == "HasAddon" then
    self:SetAuraState(aura, self:HasAddon() and "ON" or "OFF")
  end
  if aura.spellId == "NeedsInspect" then
    self:SetAuraState(aura, self:NeedsInspect() and "ON" or "OFF")
  end
end

function OrganizedPanicMember:SetAuraState(aura, state)
  if aura.state ~= state then
    aura.state = state
    NS:debug("SetAuraState", self.name, NS:ObjectToString(aura))
    AceEvent:SendMessage("AURA_STATE_CHANGE", self.name, aura.spellId, aura.state, aura.source)
  end
end
