--
-- OrganizedPanicMemberCard
--
-- Copyright (c) 2021 Dana Basken
--

local _, NS = ...

local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

OrganizedPanicMemberCard = {}
OrganizedPanicMemberCard.__index = OrganizedPanicMemberCard

local INSET = 4
local NAME_HEIGHT = 10
local MINIMUM_HEIGHT = 12 -- should match FRAME_HEIGHT in OrganizedPanicSpellCard
local NAME_WIDTH = 140
local ICON_SIZE = 12

function OrganizedPanicMemberCard:Create(member, parentFrame)
  local instance = {}
  setmetatable(instance, OrganizedPanicMemberCard)
  instance:Construct(member, parentFrame)
  return instance
end

function OrganizedPanicMemberCard:Construct(member, parentFrame)
  self.member = member
  self.parentFrame = parentFrame
  self.spellCards = {}
  self.auraIcons = {}
  self:CreateUI()
  self:RefreshUI()
  self:SetLocked(NS.db.profile.windowLocked)
  AceEvent.RegisterMessage(self, "AURA_STATE_CHANGE")
  AceEvent.RegisterMessage(self, "MEMBER_INSPECTED")
end

function OrganizedPanicMemberCard:CreateUI()
  self.frame = CreateFrame("Frame", "$parent"..OrganizedPanicCommon:GetNameWithoutRealm(self.member.name).."MemberCard", self.parentFrame, BackdropTemplateMixin and "BackdropTemplate")
  self.frame:SetWidth(self.parentFrame:GetWidth() - INSET)
  self.frame:SetHeight(0)
  self.frame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 32})
  self:SetPosition(0, 0)

  self.frame.card = CreateFrame("Frame", "$parentCard", self.frame, BackdropTemplateMixin and "BackdropTemplate")
  self.frame.card:SetWidth(self.frame:GetWidth() - (INSET * 2))
  self.frame.card:SetHeight(0)
  self.frame.card:SetPoint("TOPLEFT", self.frame, INSET, -INSET)
  self.frame.card:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 32, insets = {left = 1, right = 1, top = 1, bottom = 1}})
  self.frame.card:SetBackdropColor(0, 0, 0, 0)

  self.frame.icon = CreateFrame("Frame", "$parentIcon", self.frame, BackdropTemplateMixin and "BackdropTemplate")
  self.frame.icon:SetWidth(ICON_SIZE)
  self.frame.icon:SetHeight(ICON_SIZE)
  self.frame.icon:SetPoint("LEFT", INSET, INSET / 2)
  self.frame.icon.texture = self.frame.icon:CreateTexture()
  self.frame.icon.texture:SetAllPoints()
  self.frame.icon:Hide()

  self.frame.name = self.frame.card:CreateFontString("FontString", "OVERLAY")
  self.frame.name:SetFont("Fonts\\FRIZQT__.TTF", 10)
  self.frame.name:SetPoint("TOPLEFT", self.frame.card, ICON_SIZE + 2, 0)

  self.frame.spells = CreateFrame("Frame", "$parentSpells", self.frame.card, BackdropTemplateMixin and "BackdropTemplate")
  self.frame.spells:SetWidth(self.frame.card:GetWidth() - NAME_WIDTH)
  self.frame.spells:SetHeight(0)
  self.frame.spells:SetPoint("TOPRIGHT", self.frame.card, 0, 0)
  self.frame.spells:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 32, insets = {left = 1, right = 1, top = 1, bottom = 1}})
  self.frame.spells:SetBackdropColor(0, 0, 0, 0)
end

function OrganizedPanicMemberCard:AURA_STATE_CHANGE(event, name, spellId, state, casterGroupName)
  if name == self.member.name then self:RefreshUI() end
end

function OrganizedPanicMemberCard:MEMBER_INSPECTED(event, name)
  if name == self.member.name then self:RefreshUI() end
end

function OrganizedPanicMemberCard:FindSpell(name)
  for _, spell in pairs(self.member:GetSpells()) do
    if spell.name == name then return spell end
  end
end

function OrganizedPanicMemberCard:FindSpellCard(spell)
  for _, spellCard in pairs(self.spellCards) do
    if spellCard.spell.name == spell.name then return spellCard end
  end
end

function OrganizedPanicMemberCard:RefreshUI()
  if NS.db.profile.windowVisible then self:UpdateUI() end
end

function OrganizedPanicMemberCard:UpdateUI()
  self:UpdateIcon()
  self:UpdateName()
  self:UpdateAuras()
  local cardsHeight = self:UpdateSpellCardsUI()
  if cardsHeight == 0 then cardsHeight = MINIMUM_HEIGHT end
  self.frame:SetHeight(cardsHeight + NAME_HEIGHT)
  self.frame.card:SetHeight(self.frame:GetHeight())
  local iconY = ((ICON_SIZE) / 2) - 2
  self.frame.icon:SetPoint("TOPLEFT", 4, -iconY)
  for _, spellCard in pairs(self.spellCards) do spellCard:RefreshUI() end
end

function OrganizedPanicMemberCard:UpdateSpellCardsUI()
  self:RemoveDeadSpellCards()
  local height = 0
  for _, spell in pairs(self.member:GetSpells()) do
    if spell.minor and not NS.db.profile.includeMinorPanics then
      -- stupid lack of continue
    else
      local card = self:FindSpellCard(spell)
      if not card then
        card = OrganizedPanicSpellCard:Create(spell, self.member, self.frame.spells)
        table.insert(self.spellCards, card)
      end
      card:SetPosition(0, height)
      height = height + card:GetHeight()
    end
  end
  self.frame.spells:SetHeight(height)
  return height
end

function OrganizedPanicMemberCard:RemoveDeadSpellCards()
  for i, spellCard in pairs(self.spellCards) do
    if not self:FindSpell(spellCard.spell.name) then
      spellCard:Remove()
      self.spellCards[i] = nil
    end
  end
end

function OrganizedPanicMemberCard:UpdateIcon()
  if self.member.specializationIcon then
    self.frame.icon.texture:SetTexture(self.member.specializationIcon)
    self.frame.icon:Show()
    if NS.db.profile.windowLocked then
      self.frame.icon:SetAlpha(0.5)
    else
      self.frame.icon:SetAlpha(1)
    end
  else
    self.frame.icon:Hide()
  end
end

function OrganizedPanicMemberCard:UpdateName()
  local name = OrganizedPanicCommon:GetNameWithoutRealm(self.member.name)
  _, _, _, color = GetClassColor(self.member.class)
  self.frame.name:SetText(OrganizedPanicCommon:ColorString(name, color, true))
end

function OrganizedPanicMemberCard:UpdateAuras()
  local x = ICON_SIZE + self.frame.name:GetStringWidth() + (INSET * 2)
  local y = ((ICON_SIZE) / 2) - 2
  for _, aura in pairs(self.member:GetAuras()) do
    local card = self.auraIcons[aura.spellId]
    if card then
      card:RefreshUI()
    else
      card = OrganizedPanicAuraIcon:Create(aura, self.frame)
      self.auraIcons[aura.spellId] = card
    end
    if aura.state == "ON" then
      card:SetPosition(x, y)
      x = x + ICON_SIZE
    end
  end
end

function OrganizedPanicMemberCard:SetPosition(x, y)
  self.frame:SetPoint("TOPLEFT", self.parentFrame, x, -y)
end

function OrganizedPanicMemberCard:GetTop()
end

function OrganizedPanicMemberCard:GetHeight()
  return self.frame:GetHeight()
end

function OrganizedPanicMemberCard:GetRank()
  local bestRank = 0
  for _, spellCard in pairs(self.spellCards) do
    if spellCard.spell.minor and not NS.db.profile.includeMinorPanics then
      -- stupid lack of continue
    else
      local rank = spellCard:GetRank()
      if rank > bestRank then bestRank = rank end
    end
  end
  return bestRank
end

function OrganizedPanicMemberCard:Remove()
  NS:debug("Remove", self.member.name)
  self.frame:Hide()
  self.frame:SetParent(nil)
  self.frame:ClearAllPoints()
end

function OrganizedPanicMemberCard:SetLocked(locked)
  if locked then
    self.frame:SetBackdropColor(0.1, 0.1, 0.1, 0.5)
  else
    self.frame:SetBackdropColor(0.1, 0.1, 0.1, 0.75)
  end
  self:UpdateName()
  self:UpdateIcon()
  self.frame.icon.texture:SetAllPoints()
  for _, spellCard in pairs(self.spellCards) do spellCard:SetLocked(locked) end
  for _, auraIcon in pairs(self.auraIcons) do
    auraIcon:RefreshUI()
  end
end
