--
-- OrganizedPanicMembers
--
-- Copyright (c) 2021 Dana Basken
--

local _, NS = ...

local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

OrganizedPanicMembers = {}
OrganizedPanicMembers.__index = OrganizedPanicMembers

local PERIODIC_INSPECTIONS_INTERVAL = 1
local PERIODIC_MEMBER_CHECK_INTERVAL = 5

function OrganizedPanicMembers:Create()
  local instance = {}
  setmetatable(instance, OrganizedPanicMembers)
  instance:Construct()
  return instance
end

function OrganizedPanicMembers:Construct()
  self.members = {}
  self:SetInspecting(false)
  AceEvent.RegisterEvent(self, "INSPECT_READY")
  self:Update()
  AceTimer.ScheduleRepeatingTimer(self, "PeriodicInspections", PERIODIC_INSPECTIONS_INTERVAL)
  AceTimer.ScheduleRepeatingTimer(self, "PeriodicMemberCheck", PERIODIC_MEMBER_CHECK_INTERVAL)
end

function OrganizedPanicMembers:INSPECT_READY(event, unitId)
  local member = self:FindMemberByUnitId(unitId)
  if not member then return end
  NS:debug(event, unitId, member.name, member.groupName, member.unitName, member.groupType)
  member:Update()
  local _, specialization, _, icon, role = GetSpecializationInfoByID(GetInspectSpecialization(member:GetGroupName()))
  self:UpdateMemberSpecialization(member, specialization, icon)
  self:SetInspecting(false)
end

function OrganizedPanicMembers:SetInspecting(inspecting)
  if self.inspecting ~= inspecting then
    self.inspecting = inspecting
    self.inspectingLastChange = GetTime()
    NS:debug("SetInspecting", self.inspecting)
  end
end

function OrganizedPanicMembers:Update()
  if NS.db.profile.standalone or OrganizedPanicCommon:InGroup("player") then
    self:UpdateGroup()
    self:SetGroupedState(true)
  else
    self.members = {}
    self:SetGroupedState(false)
  end
  AceEvent:SendMessage("MEMBERS_UPDATED")
end

function OrganizedPanicMembers:SetGroupedState(grouped)
  if self.grouped ~= grouped then
    self.grouped = grouped
    AceEvent:SendMessage("GROUPED_STATE_CHANGE", self.grouped)
  end
end

function OrganizedPanicMembers:GetGroupedState() return self.grouped end

-- every second
function OrganizedPanicMembers:PeriodicInspections()
  for _, member in pairs(self:GetMembers()) do
    if member:NeedsInspect() then
      NS:debug("PeriodicInspections", "needs inspection", member.name, member.groupName)
      self:InspectMember(member)
    end
  end
  self:UpdateCooldownsRemaining()
  self:UpdateAuras()
  self:ClearLongInspections()
end

function OrganizedPanicMembers:ClearLongInspections()
  if self.inspecting and self.inspectingLastChange then
    local delta = GetTime() - self.inspectingLastChange
    NS:debug("ClearLongInspections", "inspecting last change delta", delta)
    if delta > 20 then
      self:SetInspecting(false)
    end
  end
end

-- every 5 seconds
function OrganizedPanicMembers:PeriodicMemberCheck()
  if NS.db.profile.standalone or OrganizedPanicCommon:InGroup("player") then
    self:UpdateGroup()
    self:PruneMembers()
  end
end

function OrganizedPanicMembers:PruneMembers()
  for _, member in pairs(self:GetMembers()) do
    local groupName = member:GetGroupName()
    if groupName and (not UnitIsConnected(groupName) or not self:GroupNameIsSameRole(groupName)) then
      NS:warn("PeriodicMemberCheck", "removed missing or mismatched member", member.name)
      self:DeleteMember(member)
    end
  end
end

-- every second
function OrganizedPanicMembers:UpdateCooldownsRemaining()
  for _, member in pairs(self:GetMembers()) do member:UpdateSpellCooldowns() end
end

function OrganizedPanicMembers:UpdateGroup()
  self:AddOrUpdateMember(OrganizedPanicCommon:GetPlayerNameWithRealm())
  local groupType = OrganizedPanicCommon:GetGroupType()
  if not groupType then return end
  local members = GetNumGroupMembers()
  for i = 1, members do
    local groupName = ("%s%d"):format(groupType, i)
    local name = OrganizedPanicCommon:GetUnitNameWithRealm(groupName)
    if name and UnitIsConnected(groupName) and self:GroupNameIsSameRole(groupName) then
        self:AddOrUpdateMember(name)
    else
      local member = self:FindMemberByName(name)
      if member then self:DeleteMember(member) end
    end
  end
end

function OrganizedPanicMembers:AddOrUpdateMember(name)
  local member = self:FindMemberByName(name)
  if not member then
    NS:debug("AddOrUpdateMember", name)
    member = self:AddMember(name)
  end
end

function OrganizedPanicMembers:GroupNameIsSameRole(groupName)
  if NS.db.profile.includeRole then
    local player = self:FindPlayer()
    if player and player:GetRole() ~= UnitGroupRolesAssigned(groupName) then return false end
  end
  return true
end

function OrganizedPanicMembers:AddMember(name)
  local member = OrganizedPanicMember:Create(name)
  table.insert(self.members, member)
  AceEvent:SendMessage("MEMBER_UPDATED", member.name)
  return member
end

function OrganizedPanicMembers:DeleteMember(member)
  if not member then return end
  for i, _ in pairs(self.members) do
    if self.members[i].name == member.name then
      NS:trace("DeleteMember", member.name)
      self.members[i] = nil
      AceEvent:SendMessage("MEMBER_UPDATED", member.name)
      return
    end
  end
end

function OrganizedPanicMembers:InspectMember(member)
  member:SetInspecting()
  if member:IsPlayer() then
    local _, specialization, _, icon, role = GetSpecializationInfo(GetSpecialization())
    self:UpdateMemberSpecialization(member, specialization, icon)
  else
    local groupName = member:GetGroupName()
    if not groupName or not CanInspect(groupName) then return end
    NS:debug("InspectMember", member.name, member.groupName, member.unitName, member.groupType)
    local unitName = member:GetUnitName()
    local success, error = pcall(OrganizedPanicMembers.NotifyInspect, self, unitName)
    if not success then
      NS:warn("InspectMember", "NotifyInspect failed", unitName, NS:ObjectToString(error))
    end
  end
end

function OrganizedPanicMembers:MemberInspected(member)
  AceEvent:SendMessage("MEMBER_INSPECTED", member.name, member.unitName)
  NS:debug("MemberInspected", member.name, member.groupName, GetTime() - member.lastInspectedTime)
end

function OrganizedPanicMembers:NotifyInspect(name)
  if self.inspecting then
    NS:debug("NotifyInspect", "skipping NotifyInspect", name)
  else
    NS:debug("NotifyInspect", name)
    _G["NotifyInspect"](name)
    self:SetInspecting(true)
  end
end

function OrganizedPanicMembers:UpdateMemberSpecialization(member, specialization, icon)
  if not specialization then
    NS:debug("UpdateMemberSpecialization", "blank spec", member.name)
    return
  end
  member:SetSpecialization(specialization)
  member:SetSpecializationIcon(icon)
  member:Update()
  self:MemberInspected(member)
  AceEvent:SendMessage("MEMBER_UPDATED", member.name)
end

function OrganizedPanicMembers:UpdateAuras()
  for _, member in pairs(self:GetMembers()) do member:UpdateAuras() end
end

function OrganizedPanicMembers:FindMemberByName(name)
  if not name then return end
  local nameRealm = OrganizedPanicCommon:GetNameWithRealm(name)
  for i, member in pairs(self:GetMembers()) do
    if member:GetName() == name or member:GetName() == nameRealm then return member end
  end
end

function OrganizedPanicMembers:FindMemberByUnitId(unitId)
  for i, member in pairs(self:GetMembers()) do
    if member:GetGUID() == unitId then return member end
    if unitId == "player" and member:IsPlayer() then return member end
  end
end

function OrganizedPanicMembers:FindMemberByGroupName(groupName)
  for i, member in pairs(self:GetMembers()) do
    if member:GetGroupName() == groupName then return member end
    if groupName == "player" and member:IsPlayer() then return member end
  end
end

function OrganizedPanicMembers:FindMemberByUnitName(unitName)
  return self:FindMemberByName(OrganizedPanicCommon:GetUnitNameWithRealm(unitName))
end

function OrganizedPanicMembers:FindPlayer()
  for i, member in pairs(self:GetMembers()) do
    if member:IsPlayer() then return member end
  end
end

function OrganizedPanicMembers:GetMembers()
  return self.members or {}
end
