--
-- OrganizedPanicWindow
--
-- Copyright (c) 2021 Dana Basken
--

local _, NS = ...

local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale(NS.ADDON_NAME)

OrganizedPanicWindow = {}
OrganizedPanicWindow.__index = OrganizedPanicWindow

local FRAME_WIDTH = 310
local FRAME_HEIGHT = 200
local CARD_SPACING = 1
local TITLE_HEIGHT = 42
local BUTTON_SIZE = 24
local INSET = 10
local EMPTY_POSITION = 10
local EMPTY_HEIGHT = 20

function OrganizedPanicWindow:Create(members)
  local instance = {}
  setmetatable(instance, OrganizedPanicWindow)
  instance:Construct(members)
  return instance
end

function OrganizedPanicWindow:Construct(members)
  self.members = members
  self.cards = {}
  self.needsUpdate = true
  if not NS.db.profile.windowPosition then NS.db.profile.windowPosition = {relative = "CENTER", x = 0, y = 0} end
  self:CreateUI()
  AceEvent.RegisterMessage(self, "MEMBERS_UPDATED")
  AceEvent.RegisterMessage(self, "MEMBER_UPDATED")
  AceEvent.RegisterMessage(self, "SPELL_COOLDOWN")
  AceEvent.RegisterMessage(self, "SPELL_STATE_CHANGE")
  AceTimer.ScheduleRepeatingTimer(self, "PeriodicTasks", 1)
end

function OrganizedPanicWindow:MEMBERS_UPDATED(...) self:RefreshUI() end
function OrganizedPanicWindow:MEMBER_UPDATED(...) self:RefreshUI() end
function OrganizedPanicWindow:SPELL_COOLDOWN(...) self:RefreshUI() end
function OrganizedPanicWindow:SPELL_STATE_CHANGE(...) self:RefreshUI() end

function OrganizedPanicWindow:CreateUI()
  self:CreateWindow()
  self:CreateTitleBar()
  self:CreateEmptyMessage()
  self:UpdateTitle()
  self:UpdateLockButton()
  if NS.db.profile.windowVisible then self:Show() else self:Hide() end
end

function OrganizedPanicWindow:CreateWindow()
  self.window = CreateFrame("Frame", "OrganizedPanicWindow", UIParent, BackdropTemplateMixin and "BackdropTemplate")
  self.window:SetFrameStrata("MEDIUM")
  self.window:SetWidth(FRAME_WIDTH)
  self.window:SetHeight(FRAME_HEIGHT)
  self.window:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 32, insets = {left = 0, right = 0, top = INSET, bottom = INSET}})
  self.window:SetPoint(NS.db.profile.windowPosition.relative, NS.db.profile.windowPosition.x, NS.db.profile.windowPosition.y)

  self.window:RegisterForDrag("LeftButton")
  self.window:SetScript("OnDragStart", self.window.StartMoving)
  self.window:SetScript("OnDragStop", function()
    self.window:StopMovingOrSizing()
    local point, relativeTo, relativePoint, xOfs, yOfs = self.window:GetPoint()
    NS.db.profile.windowPosition = {relative = relativePoint, x = xOfs, y = yOfs}
  end)
end

function OrganizedPanicWindow:CreateEmptyMessage()
  self.window.empty = self.window:CreateFontString("FontString", "OVERLAY")
  self.window.empty:SetFont("Fonts\\FRIZQT__.TTF", 10, "MONOCHROME")
  self.window.empty:SetPoint("CENTER", self.window, 0, -EMPTY_POSITION)
  self.window.empty:SetText(L["(not in group)"])
  self.window.empty:Hide()
end

function OrganizedPanicWindow:CreateTitleBar()
  self.titleBar = CreateFrame("Frame", "$parentTitleBar", self.window, BackdropTemplateMixin and "BackdropTemplate")
  self.titleBar:SetWidth(FRAME_WIDTH)
  self.titleBar:SetHeight(TITLE_HEIGHT)
  self.titleBar:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 32, insets = {left = 0, right = 0, top = INSET, bottom = INSET}})
  self.titleBar:SetPoint("TOPLEFT", 0, 0)
  self.titleBar:SetBackdropColor(0.1, 0, 0.1, 1)

  self.close = CreateFrame("Button","$parentClose", self.titleBar)
  self.close:SetHeight(BUTTON_SIZE)
  self.close:SetWidth(BUTTON_SIZE)
  self.close:SetPoint("TOPRIGHT", 0, -10)
  self.close:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
  self.close:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down")
  self.close:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight", "ADD")
  self.close:SetScript("OnClick", function() self:Hide() end)

  self.lock = CreateFrame("Button","$parentLock", self.titleBar)
  self.lock:SetHeight(BUTTON_SIZE)
  self.lock:SetWidth(BUTTON_SIZE)
  self.lock:SetPoint("TOPRIGHT", -20, -10)
  self.lock:SetPushedTexture("Interface\\Buttons\\LockButton-Unlocked-Down")
  self.lock:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight", "ADD")
  self.lock:SetScript("OnClick", function() self:ToggleLock() end)

  self.window.nameFrame = self.titleBar:CreateFontString("FontString", "OVERLAY")
  self.window.nameFrame:SetFont("Fonts\\FRIZQT__.TTF", 10, "MONOCHROME")
  self.window.nameFrame:SetPoint("CENTER", self.titleBar, 0, 0)
end

function OrganizedPanicWindow:PeriodicTasks()
  if self.needsUpdate then
    if NS.db.profile.windowVisible then self:UpdateUI() end
  end
end

function OrganizedPanicWindow:UpdateTitle()
  local title = NS.ADDON_TITLE
  local player = self.members:FindPlayer()
  if player and player.role and player.role ~= "NONE" then
    title = ("%s - %s"):format(title, L[player.role])
  end
  self.window.nameFrame:SetText(title)
end

function OrganizedPanicWindow:UpdateUI()
  if NS.db.profile.windowVisible then
    self:UpdateTitle()
    self:UpdateMemberCards()
    local keys = self:GetCardSortOrder()
    local y = TITLE_HEIGHT + CARD_SPACING
    if #keys == 0 then
      y = TITLE_HEIGHT + EMPTY_HEIGHT
      self.window.empty:Show()
    else
      self.window.empty:Hide()
      for i = 1, #keys do
        local card = self.cards[keys[i]]
        card:RefreshUI()
        card:SetPosition(CARD_SPACING + 1, y - INSET)
        y = y + card:GetHeight() + CARD_SPACING
      end
    end
    self.window:SetSize(self.window:GetWidth(), y + CARD_SPACING + 1)
    self.needsUpdate = false
  end
end

function OrganizedPanicWindow:GetCardSortOrder()
  local keys = {}
  for key, card in pairs(self.cards) do
    table.insert(keys, key)
  end
  table.sort(keys, function(a, b) return self.cards[a]:GetRank() < self.cards[b]:GetRank() end)
  return keys
end

function OrganizedPanicWindow:UpdateMemberCards()
  local nameMap = {}
  for _, member in pairs(self.members:GetMembers()) do
    nameMap[member.name] = member
    if not self.cards[member.name] then
      local card = OrganizedPanicMemberCard:Create(member, self.window)
      card:SetLocked(NS.db.profile.windowLocked)
      self.cards[member.name] = card
    end
  end
  for _, card in pairs(self.cards) do
    if not nameMap[card.member.name] then
      card:Remove()
      self.cards[card.member.name] = nil
    end
  end
end

function OrganizedPanicWindow:RefreshUI()
  self.needsUpdate = true
end

function OrganizedPanicWindow:Show()
  self.window:Show()
  NS.db.profile.windowVisible = true
  self:RefreshUI()
end

function OrganizedPanicWindow:Hide()
  NS.db.profile.windowVisible = false
  self.window:Hide()
end

function OrganizedPanicWindow:Toggle()
  if NS.db.profile.windowVisible then
    self:Hide()
  else
    self:Show()
  end
end

function OrganizedPanicWindow:Lock()
  NS.db.profile.windowLocked = true
  self:UpdateLockButton()
end

function OrganizedPanicWindow:Unlock()
  NS.db.profile.windowLocked = false
  self:UpdateLockButton()
end

function OrganizedPanicWindow:UpdateLockButton()
  if NS.db.profile.windowLocked then
    self.window:SetMovable(false)
    self.window:EnableMouse(false)
    self.lock:SetNormalTexture("Interface\\Buttons\\LockButton-Locked-Up")
    self.window:SetBackdropColor(0, 0, 0, 0.25)
    self.titleBar:SetBackdropColor(0.75, 0, 0.75, 0.25)
    self.window.nameFrame:SetTextColor(1, 1, 1, 0.5)
    self.window.empty:SetTextColor(1, 1, 1, 0.5)
    self.lock:SetAlpha(0.5)
    self.close:SetAlpha(0.5)
  else
    self.window:SetMovable(true)
    self.window:EnableMouse(true)
    self.lock:SetNormalTexture("Interface\\Buttons\\LockButton-Unlocked-Up")
    self.window:SetBackdropColor(0, 0, 0, 0.5)
    self.titleBar:SetBackdropColor(0.75, 0, 0.75, 0.5)
    self.window.nameFrame:SetTextColor(1, 1, 1, 1)
    self.window.empty:SetTextColor(1, 1, 1, 0.5)
    self.lock:SetAlpha(1)
    self.close:SetAlpha(1)
  end
  self:UpdateMemberCardLockedState()
end

function OrganizedPanicWindow:UpdateMemberCardLockedState()
  for _, card in pairs(self.cards) do
    card:SetLocked(NS.db.profile.windowLocked)
  end
end

function OrganizedPanicWindow:ToggleLock()
  if NS.db.profile.windowLocked then
    self:Unlock()
  else
    self:Lock()
  end
end
